use crate::dynamic_data::{DynamicLibrary, ExecutingMove, Pokemon, Script, TurnChoice};
use crate::script_implementations::rune::wrappers::*;
use crate::script_implementations::rune::RuneScriptType;
use crate::static_data::Parameter;
use crate::StringKey;
use hashbrown::HashMap;
use rune::runtime::{RuntimeContext, VmError, VmResult};
use rune::{Unit, Value};
use std::convert::TryFrom;
use std::sync::atomic::{AtomicBool, AtomicUsize};
use std::sync::Arc;

pub struct RuneScript {
    name: StringKey,
    state: Value,
    /// Returns an atomic bool for internal marking of deletion. This is currently only specifically
    /// used for deletion of a script while we are holding a reference to it (i.e. executing a script
    /// hook on it).
    marked_for_deletion: AtomicBool,
    /// A script can be suppressed by other scripts. If a script is suppressed by at least one script
    /// we will not execute its methods. This holds the number of suppressions on the script.
    suppressed_count: AtomicUsize,

    script_type: Arc<RuneScriptType>,

    runtime: Arc<RuntimeContext>,
    unit: Arc<Unit>,
}

unsafe impl Send for RuneScript {}

unsafe impl Sync for RuneScript {}

impl RuneScript {
    pub fn new(
        name: StringKey,
        object: Value,
        script_type: Arc<RuneScriptType>,
        runtime: Arc<RuntimeContext>,
        unit: Arc<Unit>,
    ) -> Self {
        Self {
            name,
            state: object,
            marked_for_deletion: Default::default(),
            suppressed_count: Default::default(),
            script_type,
            runtime,
            unit,
        }
    }

    pub(crate) fn get_state(&self) -> Value { self.state.clone() }
}

impl Script for RuneScript {
    fn name(&self) -> anyhow::Result<&StringKey> { Ok(&self.name) }

    fn get_marked_for_deletion(&self) -> &AtomicBool { &self.marked_for_deletion }

    fn get_suppressed_count(&self) -> &AtomicUsize { &self.suppressed_count }

    fn on_initialize(
        &self,
        _: &Arc<dyn DynamicLibrary>,
        pars: &HashMap<StringKey, Arc<Parameter>>,
    ) -> anyhow::Result<()> {
        if pars.is_empty() {
            return Ok(());
        }

        let mut binding = self.state.clone().into_struct()?;
        let state = binding.data_mut();
        for par in pars {
            let key = rune::alloc::string::String::try_from(par.0.str())?;
            state.insert(key, parameter_to_rune_value(par.1.as_ref())?)?;
        }
        Ok(())
    }

    fn change_speed(&self, choice: &Arc<TurnChoice>, speed: &mut u32) -> anyhow::Result<()> {
        if let Some(hash) = self.script_type.fn_change_speed {
            let mut vm = rune::runtime::Vm::new(self.runtime.clone(), self.unit.clone());
            let speed_handle = wrap_int_reference(*speed as i64)?;

            let res = vm
                .execute(
                    hash,
                    vec![
                        rune::to_value(self.state.clone())?,
                        rune::to_value(choice.wrap())?,
                        speed_handle.clone(),
                    ],
                )?
                .complete();
            if let VmResult::Err(e) = res {
                return Err(anyhow::anyhow!("Error executing script: {}", e));
            }

            *speed = get_int_reference_value(speed_handle)? as u32;
        }
        Ok(())
    }

    fn block_critical(
        &self,
        move_data: &Arc<ExecutingMove>,
        target: &Pokemon,
        hit: u8,
        block_critical: &mut bool,
    ) -> anyhow::Result<()> {
        if let Some(hash) = self.script_type.fn_block_critical {
            let mut vm = rune::runtime::Vm::new(self.runtime.clone(), self.unit.clone());
            let block_critical_handle = wrap_bool_reference(*block_critical)?;

            vm.execute(
                hash,
                vec![
                    self.state.clone(),
                    move_data.wrap(),
                    target.wrap(),
                    rune::to_value(hit)?,
                    block_critical_handle.clone(),
                ],
            )?;
            *block_critical = get_bool_reference_value(block_critical_handle)?;
        }
        Ok(())
    }

    fn as_any(&self) -> &dyn std::any::Any { self }

    fn as_any_mut(&mut self) -> &mut dyn std::any::Any { self }
}

fn parameter_to_rune_value(parameter: &Parameter) -> Result<Value, VmError> {
    match parameter {
        Parameter::Bool(b) => rune::to_value(*b),
        Parameter::Int(i) => rune::to_value(*i),
        Parameter::Float(f) => rune::to_value(*f),
        Parameter::String(s) => rune::to_value(s.str()),
    }
}
