use crate::dynamic_data::{ItemScript, Script, ScriptCategory, ScriptOwnerData, ScriptResolver};
use crate::script_implementations::rune::script::RuneScript;
use crate::script_implementations::rune::wrappers::RuneWrapper;
use crate::script_implementations::rune::RuneScriptType;
use crate::static_data::Item;
use crate::StringKey;
use hashbrown::HashMap;
use rune::compile::meta::AssociatedKind;
use rune::compile::{ComponentRef, MetaError};
use rune::diagnostics::Diagnostic;
use rune::runtime::RuntimeContext;
use rune::{Context, Diagnostics, Options, Source, Sources, Unit};
use std::any::Any;
use std::convert::TryFrom;
use std::path::Path;
use std::sync::Arc;

#[derive(Debug)]
pub struct RuneScriptResolver {
    runtime: Arc<RuntimeContext>,
    unit: Arc<Unit>,
    script_types: HashMap<(ScriptCategory, StringKey), Arc<RuneScriptType>>,
}

pub struct RuneScriptResolverBuilder {
    context: Context,
    scripts: Sources,
}

impl ScriptResolver for RuneScriptResolver {
    fn load_script(
        &self,
        owner: ScriptOwnerData,
        category: ScriptCategory,
        script_key: &StringKey,
    ) -> anyhow::Result<Option<Arc<dyn Script>>> {
        let script_type = if let Some(script_type) = self.script_types.get(&(category, script_key.clone())) {
            script_type
        } else {
            return Ok(None);
        };

        let mut o = rune::runtime::Object::new();

        let owner_obj = match owner {
            ScriptOwnerData::Pokemon(p) => Some(p.upgrade().unwrap().wrap()),
            ScriptOwnerData::BattleSide(_) => None, //FIXME
            ScriptOwnerData::Battle(_) => None,     //FIXME
            ScriptOwnerData::None => None,
        };
        if let Some(owner_obj) = owner_obj {
            o.insert(rune::alloc::String::try_from("owner")?, owner_obj.into())?;
        }

        let state = rune::to_value(o)?;
        let script = Arc::new(RuneScript::new(
            script_key.clone(),
            state,
            script_type.clone(),
            self.runtime.clone(),
            self.unit.clone(),
        ));
        Ok(Some(script))
    }

    fn load_item_script(&self, _key: &dyn Item) -> anyhow::Result<Option<Arc<dyn ItemScript>>> { Ok(None) }

    fn as_any(&self) -> &dyn Any { self }
}

impl RuneScriptResolverBuilder {
    pub fn new() -> anyhow::Result<Self> {
        let mut context = Context::with_default_modules()?;
        context.install(super::wrappers::module()?)?;

        Ok(Self {
            context,
            scripts: Sources::default(),
        })
    }

    pub fn insert_script(&mut self, path: &Path, name: &str, script: &str) -> anyhow::Result<&mut Self> {
        self.scripts
            .insert(Source::with_path(name, script.to_string(), path)?)?;
        Ok(self)
    }

    pub fn build(mut self) -> anyhow::Result<Arc<dyn ScriptResolver>> {
        let mut visitor = FindScriptTypeVisitor::default();
        let mut diagnostics = Diagnostics::new();

        let mut options = Options::default();
        options.debug_info(true);
        options.memoize_instance_fn(true);
        options.macros(true);
        options.bytecode(true);

        let result = rune::prepare(&mut self.scripts)
            .with_context(&self.context)
            .with_visitor(&mut visitor)?
            .with_diagnostics(&mut diagnostics)
            .with_options(&options)
            .build();
        if diagnostics.has_error() {
            let error_message = diagnostics
                .diagnostics()
                .iter()
                .filter_map(|d| match d {
                    Diagnostic::Fatal(f) => Some(f.to_string()),
                    _ => None,
                })
                .collect::<Vec<String>>()
                .join("\n");

            return Err(anyhow::anyhow!("Error building Rune script: {}", error_message));
        }

        let mut script_types = HashMap::with_capacity(visitor.script_types.len());
        for (key, script_type) in visitor.script_types {
            script_types.insert(key, Arc::new(script_type));
        }

        Ok(Arc::new(RuneScriptResolver {
            runtime: Arc::new(self.context.runtime()?),
            unit: Arc::new(result?),
            script_types,
        }))
    }
}

#[derive(Debug, Default)]
struct FindScriptTypeVisitor {
    script_types: HashMap<(ScriptCategory, StringKey), RuneScriptType>,
}

impl rune::compile::CompileVisitor for FindScriptTypeVisitor {
    fn register_meta(&mut self, meta: rune::compile::MetaRef<'_>) -> Result<(), MetaError> {
        match meta.kind {
            rune::compile::meta::Kind::Struct { .. } => {
                if meta.item.iter().count() < 2 {
                    return Ok(());
                }
                #[allow(clippy::unwrap_used)] // We know that the first element exists
                let mod_name = meta.item.iter().nth(0).unwrap();
                let category = match get_mod_category(mod_name) {
                    Ok(value) => value,
                    Err(value) => return value,
                };
                #[allow(clippy::unwrap_used)] // We know that the last element exists
                let name = meta.item.last().unwrap();
                self.script_types
                    .insert((category, name.to_string().as_str().into()), RuneScriptType::default());
            }
            rune::compile::meta::Kind::Function {
                associated: Some(AssociatedKind::Instance(associated)),
                ..
            } => {
                if meta.item.iter().count() < 3 {
                    return Ok(());
                }
                let mod_name = meta.item.iter().nth(0).unwrap();
                let category = match get_mod_category(mod_name) {
                    Ok(value) => value,
                    Err(value) => return value,
                };
                let instance = meta.item.iter().nth_back(1).unwrap();
                if let Some(script_type) = self
                    .script_types
                    .get_mut(&(category, instance.to_string().as_str().into()))
                {
                    script_type.on_found_fn(associated.to_string().as_str(), meta.hash);
                }
            }
            _ => {}
        }

        Ok(())
    }
}

fn get_mod_category(mod_name: ComponentRef) -> Result<ScriptCategory, Result<(), MetaError>> {
    Ok(match mod_name.to_string().as_str() {
        "moves" => ScriptCategory::Move,
        "abilities" => ScriptCategory::Ability,
        "status" => ScriptCategory::Status,
        "pokemon" => ScriptCategory::Pokemon,
        "sides" => ScriptCategory::Side,
        "battle" => ScriptCategory::Battle,
        "weather" => ScriptCategory::Weather,
        "item_battle_triggers" => ScriptCategory::ItemBattleTrigger,
        _ => return Err(Ok(())),
    })
}
