use crate::dynamic_data::TurnChoice;
use crate::script_implementations::rune::wrappers::RuneWrapper;
use rune::runtime::Value;
use rune::Any;
use std::sync::Arc;

pub fn register(module: &mut rune::Module) -> anyhow::Result<()> {
    module.ty::<RuneTurnChoice>()?;
    module.function_meta(RuneTurnChoice::user)?;
    module.function_meta(RuneTurnChoice::speed)?;
    module.function_meta(RuneTurnChoice::has_failed)?;
    module.function_meta(RuneTurnChoice::fail)?;

    module.ty::<RuneMoveChoice>()?;
    module.function_meta(RuneMoveChoice::used_move)?;
    module.function_meta(RuneMoveChoice::target_side)?;
    module.function_meta(RuneMoveChoice::target_index)?;
    module.function_meta(RuneMoveChoice::priority)?;
    module.function_meta(RuneMoveChoice::user)?;

    module.ty::<RuneItemChoice>()?;
    module.ty::<RuneSwitchChoice>()?;
    module.ty::<RuneFleeChoice>()?;
    module.ty::<RunePassChoice>()?;

    Ok(())
}

#[derive(Debug, Clone, Any)]
pub enum RuneTurnChoice {
    Move(#[rune(get)] RuneMoveChoice),
    Item(#[rune(get)] RuneItemChoice),
    Switch(#[rune(get)] RuneSwitchChoice),
    Flee(#[rune(get)] RuneFleeChoice),
    Pass(#[rune(get)] RunePassChoice),
}

#[derive(Debug, Clone, Any)]
pub struct RuneMoveChoice(Arc<TurnChoice>);

#[derive(Debug, Clone, Any)]
pub struct RuneItemChoice(Arc<TurnChoice>);

#[derive(Debug, Clone, Any)]
pub struct RuneSwitchChoice(Arc<TurnChoice>);

#[derive(Debug, Clone, Any)]
pub struct RuneFleeChoice(Arc<TurnChoice>);

#[derive(Debug, Clone, Any)]
pub struct RunePassChoice(Arc<TurnChoice>);

impl RuneTurnChoice {
    pub fn get_turn_choice(&self) -> &Arc<TurnChoice> {
        match self {
            RuneTurnChoice::Move(m) => &m.0,
            RuneTurnChoice::Item(i) => &i.0,
            RuneTurnChoice::Switch(s) => &s.0,
            RuneTurnChoice::Flee(f) => &f.0,
            RuneTurnChoice::Pass(p) => &p.0,
        }
    }

    #[rune::function]
    fn user(&self) -> Value { self.get_turn_choice().user().wrap() }

    #[rune::function]
    fn speed(&self) -> u32 { self.get_turn_choice().speed() }

    #[rune::function]
    fn has_failed(&self) -> bool { self.get_turn_choice().has_failed() }

    #[rune::function]
    fn fail(&self) { self.get_turn_choice().fail() }
}

impl RuneWrapper for &Arc<TurnChoice> {
    fn wrap(self) -> Value {
        let o = match self.as_ref() {
            TurnChoice::Move(_) => RuneTurnChoice::Move(RuneMoveChoice(self.clone())),
            TurnChoice::Item(_) => RuneTurnChoice::Item(RuneItemChoice(self.clone())),
            TurnChoice::Switch(_) => RuneTurnChoice::Switch(RuneSwitchChoice(self.clone())),
            TurnChoice::Flee(_) => RuneTurnChoice::Flee(RuneFleeChoice(self.clone())),
            TurnChoice::Pass(_) => RuneTurnChoice::Pass(RunePassChoice(self.clone())),
        };
        rune::to_value(o).unwrap()
    }
}

impl RuneMoveChoice {
    fn move_choice(&self) -> &crate::dynamic_data::MoveChoice {
        match self.0.as_ref() {
            TurnChoice::Move(m) => m,
            _ => unreachable!("RuneMoveChoice should only be created with a MoveChoice"),
        }
    }

    #[rune::function]
    fn used_move(&self) -> Value { self.move_choice().used_move().wrap() }

    #[rune::function]
    fn target_side(&self) -> u8 { self.move_choice().target_side() }

    #[rune::function]
    fn target_index(&self) -> u8 { self.move_choice().target_index() }

    #[rune::function]
    fn priority(&self) -> i8 { self.move_choice().priority() }

    #[rune::function]
    fn user(&self) -> Value { self.move_choice().user().wrap() }
}
