use rune::alloc::fmt::TryWrite;
use rune::runtime::{Protocol, VmResult};
use rune::{Any, Value};
use std::num::Saturating;

mod dynamic_data;
mod parameters;
mod static_data;

pub trait RuneWrapper {
    fn wrap(self) -> Value;
}

pub fn module() -> anyhow::Result<rune::Module> {
    let mut module = rune::Module::new();
    module.ty::<RuneValueIntWrapper>()?;
    module.function_meta(RuneValueIntWrapper::rn_as_int)?;
    module.function_meta(RuneValueIntWrapper::set_value)?;
    module.associated_function(Protocol::ADD_ASSIGN, RuneValueIntWrapper::add_assign)?;
    module.associated_function(Protocol::SUB_ASSIGN, RuneValueIntWrapper::sub_assign)?;
    module.associated_function(Protocol::DIV_ASSIGN, RuneValueIntWrapper::div_assign)?;
    module.associated_function(Protocol::MUL_ASSIGN, RuneValueIntWrapper::mul_assign)?;
    module.associated_function(Protocol::PARTIAL_EQ, RuneValueIntWrapper::partial_eq)?;
    module.associated_function(Protocol::EQ, RuneValueIntWrapper::eq)?;
    module.function_meta(RuneValueIntWrapper::string_display)?;

    module.ty::<RuneValueBoolWrapper>()?;
    module.function_meta(RuneValueBoolWrapper::rn_as_bool)?;
    module.function_meta(RuneValueBoolWrapper::set_value)?;
    module.associated_function(Protocol::PARTIAL_EQ, RuneValueBoolWrapper::eq)?;
    module.associated_function(Protocol::EQ, RuneValueBoolWrapper::eq)?;
    module.function_meta(RuneValueBoolWrapper::string_display)?;

    module.ty::<RuneStringKey>()?;
    module.function_meta(RuneStringKey::string_display)?;
    module.associated_function(Protocol::PARTIAL_EQ, RuneStringKey::eq)?;
    module.associated_function(Protocol::EQ, RuneStringKey::eq)?;

    parameters::register(&mut module)?;
    dynamic_data::register(&mut module)?;
    static_data::register(&mut module)?;
    Ok(module)
}

pub fn wrap_int_reference(value: i64) -> anyhow::Result<Value> { Ok(rune::to_value(RuneValueIntWrapper::new(value))?) }

pub fn wrap_bool_reference(value: bool) -> anyhow::Result<Value> {
    Ok(rune::to_value(RuneValueBoolWrapper::new(value))?)
}

pub fn get_int_reference_value(value: Value) -> anyhow::Result<i64> {
    let obj = match value.into_any::<RuneValueIntWrapper>() {
        Ok(obj) => obj,
        Err(_) => return Err(anyhow::anyhow!("Value is not a RuneValueIntWrapper")),
    };
    Ok(obj.as_int())
}

pub fn get_bool_reference_value(value: Value) -> anyhow::Result<bool> {
    let obj = match value.into_any::<RuneValueBoolWrapper>() {
        Ok(obj) => obj,
        Err(_) => return Err(anyhow::anyhow!("Value is not a RuneValueBoolWrapper")),
    };
    Ok(obj.as_bool())
}

macro_rules! impl_rune_wrapper {
    ($t:ty, $wrapped_type:ident) => {
        impl crate::script_implementations::rune::wrappers::RuneWrapper for $t {
            fn wrap(self) -> rune::runtime::Value { rune::to_value($wrapped_type(self.clone())).unwrap() }
        }
    };
}

use crate::StringKey;
use impl_rune_wrapper;

#[derive(Any, Clone)]
#[rune(name = RefInt)]
struct RuneValueIntWrapper {
    value: Saturating<i64>,
}

impl RuneValueIntWrapper {
    pub fn new(value: i64) -> Self {
        Self {
            value: Saturating(value),
        }
    }

    pub fn as_int(&self) -> i64 { self.value.0 }

    #[rune::function(path = RuneValueIntWrapper::as_int)]
    fn rn_as_int(&self) -> i64 { self.value.0 }

    #[rune::function]
    fn set_value(&mut self, value: i64) { self.value.0 = value; }

    fn add_assign(&mut self, other: i64) { self.value += other; }

    fn sub_assign(&mut self, other: i64) { self.value -= other; }

    fn div_assign(&mut self, other: Value) {
        match other.as_integer() {
            Ok(other) => self.value /= other,
            Err(_) => match other.as_float() {
                Ok(other) => self.div_assign_f64(other),
                Err(_) => (),
            },
        }
    }

    fn div_assign_f64(&mut self, other: f64) {
        let v = (self.value.0 as f64 / other).floor();
        if v > i64::MAX as f64 {
            self.value = Saturating(i64::MAX);
        } else if v < i64::MIN as f64 {
            self.value = Saturating(i64::MIN);
        } else {
            self.value = Saturating(v as i64);
        }
    }

    fn mul_assign(&mut self, other: Value) {
        match other.as_integer() {
            Ok(other) => self.value *= other,
            Err(_) => match other.as_float() {
                Ok(other) => self.mul_assign_f64(other),
                Err(_) => (),
            },
        }
    }

    fn mul_assign_f64(&mut self, other: f64) {
        let v = (self.value.0 as f64 * other).floor();
        if v > i64::MAX as f64 {
            self.value = Saturating(i64::MAX);
        } else if v < i64::MIN as f64 {
            self.value = Saturating(i64::MIN);
        } else {
            self.value = Saturating(v as i64);
        }
    }

    fn partial_eq(&self, b: i64) -> VmResult<bool> { VmResult::Ok(self.value.0 == b) }

    fn eq(&self, b: i64) -> VmResult<bool> { VmResult::Ok(self.value.0 == b) }

    #[rune::function(instance, protocol = STRING_DISPLAY)]
    fn string_display(&self, f: &mut rune::runtime::Formatter) -> VmResult<()> {
        rune::vm_write!(f, "{}", self.value.0);
        VmResult::Ok(())
    }
}

#[derive(Any, Clone)]
#[rune(name = RefBool)]
struct RuneValueBoolWrapper(bool);

impl RuneValueBoolWrapper {
    pub fn new(value: bool) -> Self { Self(value) }

    pub fn as_bool(&self) -> bool { self.0 }

    #[rune::function(path = RuneValueIntWrapper::as_bool)]
    fn rn_as_bool(&self) -> bool { self.0 }

    #[rune::function]
    fn set_value(&mut self, value: bool) { self.0 = value; }

    #[rune::function(instance, protocol = STRING_DISPLAY)]
    fn string_display(&self, f: &mut rune::runtime::Formatter) -> VmResult<()> {
        rune::vm_write!(f, "{}", self.0);
        VmResult::Ok(())
    }

    fn eq(&self, b: bool) -> VmResult<bool> { VmResult::Ok(self.0 == b) }
}

#[derive(Any, Clone, Debug)]
#[rune(name = StringKey)]
pub(super) struct RuneStringKey(pub StringKey);

impl RuneStringKey {
    pub fn new(value: StringKey) -> Self { Self(value) }

    #[rune::function(instance, protocol = STRING_DISPLAY)]
    fn string_display(&self, f: &mut rune::runtime::Formatter) -> VmResult<()> {
        rune::vm_write!(f, "{}", self.0);
        VmResult::Ok(())
    }

    fn eq(&self, b: Value) -> VmResult<bool> {
        match b.borrow_string_ref() {
            Ok(s) => VmResult::Ok(self.0 == StringKey::new(&s)),
            Err(_) => {
                let b = b.borrow_any_ref::<RuneStringKey>();
                match b {
                    Ok(b) => VmResult::Ok(self.0 == b.0),
                    _ => VmResult::Ok(false),
                }
            }
        }
    }
}

impl RuneWrapper for &StringKey {
    fn wrap(self) -> Value { rune::to_value(RuneStringKey(self.clone())).unwrap() }
}

#[cfg(test)]
mod tests {
    use super::*;
    use anyhow::Result;
    use rune::diagnostics::Diagnostic;
    use rune::{Context, Diagnostics, Options, Source, Vm};
    use std::sync::Arc;

    pub fn setup_script(script: &str) -> Result<Vm> {
        let mut context = Context::with_default_modules()?;
        context.install(module()?)?;

        let mut sources = rune::Sources::new();
        sources.insert(Source::memory(script)?)?;
        let mut diagnostics = Diagnostics::new();

        let mut options = Options::default();
        options.debug_info(true);
        options.memoize_instance_fn(true);
        options.macros(true);
        options.bytecode(true);

        let unit = rune::prepare(&mut sources)
            .with_context(&context)
            .with_diagnostics(&mut diagnostics)
            .with_options(&options)
            .build();
        if !diagnostics.is_empty() && diagnostics.has_error() {
            let error_message = diagnostics
                .diagnostics()
                .iter()
                .filter_map(|d| match d {
                    Diagnostic::Fatal(f) => Some(f.to_string()),
                    _ => None,
                })
                .collect::<Vec<String>>()
                .join("\n");

            return Err(anyhow::anyhow!("Error building Rune script: {}", error_message));
        }

        Ok(Vm::new(Arc::new(context.runtime()?), Arc::new(unit?)))
    }

    macro_rules! execute_vm {
        ($vm:expr, $func:expr, $val:expr) => {{
            let args = vec![$val.clone()];
            $vm.execute([$func], args)?.complete().into_result()?
        }};
    }

    pub(crate) use execute_vm;

    #[test]
    fn test_int_wrapper_set_value() -> Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_int(a) {
                a.set_value(5);
            }
            "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 5);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_as_int() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a.as_int()
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        let res = execute_vm!(vm, "test_int", val);
        let res = res.as_integer()?;
        assert_eq!(res, 10);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_add() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a += 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 15);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_add_overflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a += 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MAX)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MAX);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_sub() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a -= 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 5);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_underflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a -= 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MIN)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MIN);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_mul() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a *= 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 50);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_mul_float() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a *= 0.5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 5);
        Ok(())
    }
    #[test]
    fn test_int_wrapper_mul_overflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a *= 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MAX)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MAX);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_mul_float_overflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a *= 10.0;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MAX)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MAX);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_mul_float_underflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a *= 10.0;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MIN)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MIN);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_div() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a /= 5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 2);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_div_float() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a /= 0.5;
        }
        "#,
        )?;
        let val = wrap_int_reference(10)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, 20);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_div_float_overflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a /= 0.0001;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MAX)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MAX);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_div_float_underflow() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a /= 0.0001;
        }
        "#,
        )?;
        let val = wrap_int_reference(i64::MIN)?;
        execute_vm!(vm, "test_int", val);
        let v = get_int_reference_value(val)?;
        assert_eq!(v, i64::MIN);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_eq() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a == 5
        }
        "#,
        )?;
        let val = wrap_int_reference(5)?;
        let res = execute_vm!(vm, "test_int", val);
        let res = res.as_bool()?;
        assert_eq!(res, true);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_ineq() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            a == 5
        }
        "#,
        )?;
        let val = wrap_int_reference(6)?;
        let res = execute_vm!(vm, "test_int", val);
        let res = res.as_bool()?;
        assert_eq!(res, false);
        Ok(())
    }

    #[test]
    fn test_int_wrapper_string_display() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_int(a) {
            `${a}`
        }
        "#,
        )?;
        let val = wrap_int_reference(5)?;
        let res = execute_vm!(vm, "test_int", val);
        let res = res.into_string()?.into_std();
        assert_eq!(res, "5");
        Ok(())
    }

    #[test]
    fn test_bool_wrapper_set_value() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_bool(a) {
            a.set_value(true);
        }
        "#,
        )?;
        let val = wrap_bool_reference(false)?;
        execute_vm!(vm, "test_bool", val);
        let v = get_bool_reference_value(val)?;
        assert_eq!(v, true);
        Ok(())
    }

    #[test]
    fn test_bool_wrapper_as_bool() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_bool(a) {
            a.as_bool()
        }
        "#,
        )?;
        let val = wrap_bool_reference(true)?;
        let res = execute_vm!(vm, "test_bool", val);
        let res = res.as_bool()?;
        assert_eq!(res, true);
        Ok(())
    }

    #[test]
    fn test_bool_wrapper_eq() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_bool(a) {
            a == true
        }
        "#,
        )?;
        let val = wrap_bool_reference(true)?;
        let res = execute_vm!(vm, "test_bool", val);
        let res = res.as_bool()?;
        assert_eq!(res, true);
        Ok(())
    }

    #[test]
    fn test_bool_wrapper_ineq() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_bool(a) {
            a == true
        }
        "#,
        )?;
        let val = wrap_bool_reference(false)?;
        let res = execute_vm!(vm, "test_bool", val);
        let res = res.as_bool()?;
        assert_eq!(res, false);
        Ok(())
    }

    #[test]
    fn test_bool_wrapper_string_display() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_bool(a) {
            `${a}`
        }
        "#,
        )?;
        let val = wrap_bool_reference(true)?;
        let res = execute_vm!(vm, "test_bool", val);
        let res = res.into_string()?.into_std();
        assert_eq!(res, "true");
        Ok(())
    }

    #[test]
    fn test_string_key_wrapper_string_display() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_string_key(a) {
            `${a}`
        }
        "#,
        )?;
        let val = StringKey::from("test");
        let arg = val.wrap();
        let res = execute_vm!(vm, "test_string_key", arg);
        let res = res.into_string()?.into_std();
        assert_eq!(res, "test");
        Ok(())
    }

    #[test]
    fn test_string_key_wrapper_eq() -> Result<()> {
        let mut vm = setup_script(
            r#"
        pub fn test_string_key(a) {
            a == "test"
        }
        "#,
        )?;
        let val = StringKey::from("test");
        let arg = val.wrap();
        let res = execute_vm!(vm, "test_string_key", arg);
        let res = res.as_bool()?;
        assert_eq!(res, true);
        Ok(())
    }
}
