use crate::script_implementations::rune::wrappers::{impl_rune_wrapper, RuneStringKey, RuneWrapper};
use crate::static_data::Ability;
use crate::StringKey;
use rune::runtime::Value;
use rune::Any;
use std::sync::Arc;

pub fn register(module: &mut rune::Module) -> anyhow::Result<()> {
    module.ty::<RuneAbility>()?;
    module.function_meta(RuneAbility::name)?;
    module.function_meta(RuneAbility::effect)?;
    module.function_meta(RuneAbility::get_parameter)?;
    Ok(())
}

#[derive(Debug, Any)]
pub struct RuneAbility(Arc<dyn Ability>);

impl_rune_wrapper!(&Arc<dyn Ability>, RuneAbility);

impl RuneAbility {
    #[rune::function]
    fn name(&self) -> Value { self.0.name().wrap() }

    #[rune::function]
    fn effect(&self) -> Value { self.0.effect().wrap() }

    #[rune::function]
    fn get_parameter(&self, key: Value) -> Option<Value> {
        if let Ok(s) = key.borrow_string_ref() {
            return self.0.parameters().get(&StringKey::new(&s)).map(|v| v.wrap());
        }
        if let Ok(v) = key.into_any::<RuneStringKey>() {
            self.0.parameters().get(&v.0).map(|v| v.wrap())
        } else {
            None
        }
    }
}

#[cfg(test)]
mod tests {
    use super::super::super::tests::{execute_vm, setup_script};
    use super::*;
    use crate::static_data::tests::MockAbility;
    use crate::static_data::Parameter;
    use hashbrown::HashMap;

    #[test]
    fn test_ability_name() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_ability(ability) {
                let name = ability.name();
                assert_eq!(name, "TestAbility");
            }
            "#,
        )?;
        let mut ability = MockAbility::new();
        ability.expect_name().once().return_const(StringKey::new("TestAbility"));
        let ability: Arc<dyn Ability> = Arc::new(ability);
        let wrapped = ability.wrap();
        execute_vm!(&mut vm, "test_ability", wrapped);
        Ok(())
    }

    #[test]
    fn test_ability_effect() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_ability(ability) {
                let effect = ability.effect();
                assert_eq!(effect, "TestEffect");
            }
            "#,
        )?;
        let mut ability = MockAbility::new();
        ability
            .expect_effect()
            .once()
            .return_const(StringKey::new("TestEffect"));
        let ability: Arc<dyn Ability> = Arc::new(ability);
        let wrapped = ability.wrap();
        execute_vm!(&mut vm, "test_ability", wrapped);
        Ok(())
    }

    #[test]
    fn test_ability_get_parameter() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_ability(ability) {
                let value = ability.get_parameter("test_key");
                match value {
                    Some(v) => {
                        match v {
                            Parameter::String(s) => assert_eq!(s, "TestValue"),
                            _ => panic!("Invalid value type"),
                        }
                    }
                    None => panic!("Value not found"),
                }
            }
            "#,
        )?;
        let mut ability = MockAbility::new();
        ability.expect_parameters().once().return_const({
            let mut map = HashMap::new();
            map.insert(
                StringKey::new("test_key"),
                Arc::new(Parameter::String(StringKey::new("TestValue"))),
            );
            map
        });
        let a: Arc<dyn Ability> = Arc::new(ability);
        let wrapped = a.wrap();
        execute_vm!(&mut vm, "test_ability", wrapped);
        Ok(())
    }
}
