use crate::script_implementations::rune::wrappers::{impl_rune_wrapper, RuneStringKey, RuneWrapper};
use crate::static_data::{AbilityIndex, Form, Statistic};
use crate::StringKey;
use rune::runtime::Value;
use rune::Any;
use std::sync::Arc;

pub fn register(module: &mut rune::Module) -> anyhow::Result<()> {
    module.ty::<RuneForm>()?;
    module.function_meta(RuneForm::name)?;
    module.function_meta(RuneForm::height)?;
    module.function_meta(RuneForm::weight)?;
    module.function_meta(RuneForm::base_experience)?;
    module.function_meta(RuneForm::types)?;
    module.function_meta(RuneForm::base_stats)?;
    module.function_meta(RuneForm::abilities)?;
    module.function_meta(RuneForm::hidden_abilities)?;
    module.function_meta(RuneForm::has_flag)?;
    module.function_meta(RuneForm::get_type)?;
    module.function_meta(RuneForm::get_base_stat)?;
    module.function_meta(RuneForm::get_ability)?;
    Ok(())
}

#[derive(Debug, Any)]
#[rune(name = Form)]
pub struct RuneForm(pub Arc<dyn Form>);

impl_rune_wrapper!(&Arc<dyn Form>, RuneForm);

impl RuneForm {
    #[rune::function]
    fn name(&self) -> Value { self.0.name().wrap() }

    #[rune::function]
    fn height(&self) -> f32 { self.0.height() }

    #[rune::function]
    fn weight(&self) -> f32 { self.0.weight() }

    #[rune::function]
    fn base_experience(&self) -> u32 { self.0.base_experience() }

    #[rune::function]
    fn types(&self) -> Vec<i64> { self.0.types().iter().map(|t| u8::from(*t) as i64).collect() }

    #[rune::function]
    fn base_stats(&self) -> Value { self.0.base_stats().wrap() }

    #[rune::function]
    fn abilities(&self) -> Vec<Value> { self.0.abilities().iter().map(|a| a.wrap()).collect() }

    #[rune::function]
    fn hidden_abilities(&self) -> Vec<Value> { self.0.hidden_abilities().iter().map(|a| a.wrap()).collect() }

    #[rune::function]
    fn has_flag(&self, key: Value) -> bool {
        if let Ok(s) = key.borrow_string_ref() {
            return self.0.has_flag(&StringKey::new(&s));
        }
        if let Ok(v) = key.into_any::<RuneStringKey>() {
            return self.0.has_flag(&v.0);
        }
        false
    }

    #[rune::function]
    fn get_type(&self, index: usize) -> anyhow::Result<i64> { self.0.get_type(index).map(|t| u8::from(t) as i64) }

    #[rune::function]
    fn get_base_stat(&self, statistic: Statistic) -> u16 { self.0.get_base_stat(statistic) }

    #[rune::function]
    fn get_ability(&self, hidden: bool, index: i64) -> anyhow::Result<Value> {
        if index > u8::MAX as i64 || index < 0 {
            return Err(anyhow::anyhow!("Index out of bounds"));
        }
        let index = index as u8;
        self.0.get_ability(AbilityIndex { hidden, index }).map(|a| a.wrap())
    }
}

#[cfg(test)]
mod tests {
    use super::super::super::tests::{execute_vm, setup_script};
    use super::*;
    use crate::static_data::tests::MockForm;
    use crate::static_data::{StaticStatisticSet, TypeIdentifier};

    #[test]
    fn test_form_name() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let name = form.name();
                assert_eq!(name, "TestForm");
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_name().once().return_const(StringKey::new("TestForm"));
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_height() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let height = form.height();
                assert_eq!(height, 1.0);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_height().once().return_const(1.0);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_weight() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let weight = form.weight();
                assert_eq!(weight, 1.0);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_weight().once().return_const(1.0);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_base_experience() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let base_experience = form.base_experience();
                assert_eq!(base_experience, 1);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_base_experience().once().return_const(1u32);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_types() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let types = form.types();
                assert_eq!(types[0], 1);
                assert_eq!(types[1], 2);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_types()
            .once()
            .return_const(vec![TypeIdentifier::from(1u8), TypeIdentifier::from(2u8)]);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_base_stats() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let base_stats = form.base_stats();
                assert_eq!(base_stats.hp(), 1);
                assert_eq!(base_stats.attack(), 2);
                assert_eq!(base_stats.defense(), 3);
                assert_eq!(base_stats.special_attack(), 4);
                assert_eq!(base_stats.special_defense(), 5);
                assert_eq!(base_stats.speed(), 6);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_base_stats()
            .once()
            .return_const(Arc::new(StaticStatisticSet::<u16>::new(1, 2, 3, 4, 5, 6)));
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_abilities() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let abilities = form.abilities();
                assert_eq!(abilities.len(), 1);
                assert_eq!(abilities[0], "TestAbility");
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_abilities()
            .once()
            .return_const(vec![StringKey::new("TestAbility")]);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_hidden_abilities() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let abilities = form.hidden_abilities();
                assert_eq!(abilities.len(), 1);
                assert_eq!(abilities[0], "TestAbility");
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_hidden_abilities()
            .once()
            .return_const(vec![StringKey::new("TestAbility")]);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_has_flag() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let has_flag = form.has_flag("test_key");
                assert_eq!(has_flag, true);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_has_flag().once().return_const(true);
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_get_type() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let type_id = form.get_type(0)?;
                assert_eq!(type_id, 1);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_get_type()
            .once()
            .returning(|_| Ok(TypeIdentifier::from(1u8)));
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_get_base_stat() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let base_stat = form.get_base_stat(Statistic::HP);
                assert_eq!(base_stat, 1);
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_get_base_stat()
            .once()
            .return_const(StaticStatisticSet::<u16>::new(1, 2, 3, 4, 5, 6).hp());
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }

    #[test]
    fn test_form_get_ability() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_form(form) {
                let ability = form.get_ability(false, 0)?;
                assert_eq!(ability, "TestAbility");
            }
            "#,
        )?;
        let mut form = MockForm::new();
        form.expect_get_ability()
            .once()
            .returning(move |_| Ok(StringKey::new("TestAbility")));
        let form: Arc<dyn Form> = Arc::new(form);
        let wrapped = form.wrap();
        execute_vm!(&mut vm, "test_form", wrapped);
        Ok(())
    }
}
