use crate::defines::LevelInt;
use crate::script_implementations::rune::wrappers::impl_rune_wrapper;
use crate::static_data::GrowthRate;
use rune::Any;
use std::sync::Arc;

pub fn register(module: &mut rune::Module) -> anyhow::Result<()> {
    module.ty::<RuneGrowthRate>()?;
    module.function_meta(RuneGrowthRate::calculate_level)?;
    module.function_meta(RuneGrowthRate::calculate_experience)?;
    Ok(())
}

#[derive(Debug, Any)]
pub struct RuneGrowthRate(Arc<dyn GrowthRate>);

impl_rune_wrapper!(&Arc<dyn GrowthRate>, RuneGrowthRate);

impl RuneGrowthRate {
    #[rune::function]
    fn calculate_level(&self, experience: u32) -> i32 { self.0.calculate_level(experience) as i32 }

    #[rune::function]
    fn calculate_experience(&self, level: i64) -> Result<u32, String> {
        if level < 0 {
            return Err("Level cannot be negative".to_string());
        }
        if level > LevelInt::MAX as i64 {
            return Err(format!("Level cannot be greater than {}", LevelInt::MAX));
        }
        let level = level as LevelInt;

        self.0.calculate_experience(level).map_err(|e| e.to_string())
    }
}

#[cfg(test)]
mod tests {
    use super::super::super::tests::{execute_vm, setup_script};
    use super::*;
    use crate::script_implementations::rune::wrappers::RuneWrapper;
    use crate::static_data::tests::MockGrowthRate;

    #[test]
    fn test_calculate_level() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_growth_rate(growth_rate) {
                let level = growth_rate.calculate_level(10);
                assert_eq!(level, 10);
            }
            "#,
        )?;
        let mut growth_rate = MockGrowthRate::new();
        growth_rate.expect_calculate_level().returning(|_| 10);
        let growth_rate: Arc<dyn GrowthRate> = Arc::new(growth_rate);
        let wrapped = growth_rate.wrap();
        execute_vm!(&mut vm, "test_growth_rate", wrapped);
        Ok(())
    }

    #[test]
    fn test_calculate_experience() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_growth_rate(growth_rate) {
                let experience = growth_rate.calculate_experience(10)?;
                assert_eq!(experience, 10);
            }
            "#,
        )?;
        let mut growth_rate = MockGrowthRate::new();
        growth_rate.expect_calculate_experience().returning(|_| Ok(10));
        let growth_rate: Arc<dyn GrowthRate> = Arc::new(growth_rate);
        let wrapped = growth_rate.wrap();
        execute_vm!(&mut vm, "test_growth_rate", wrapped);
        Ok(())
    }
}
