use std::sync::Arc;

use rune::runtime::Value;
use rune::Any;

use crate::script_implementations::rune::wrappers::{impl_rune_wrapper, RuneStringKey, RuneWrapper};
use crate::static_data::{BattleItemCategory, Item, ItemCategory};

pub fn register(module: &mut rune::Module) -> anyhow::Result<()> {
    module.ty::<ItemCategory>()?;
    module.ty::<BattleItemCategory>()?;

    module.ty::<RuneItem>()?;
    module.function_meta(RuneItem::name)?;
    module.function_meta(RuneItem::category)?;
    module.function_meta(RuneItem::battle_category)?;
    module.function_meta(RuneItem::price)?;
    module.function_meta(RuneItem::flags)?;
    module.function_meta(RuneItem::has_flag)?;

    Ok(())
}

#[derive(Debug, Any)]
pub struct RuneItem(pub Arc<dyn Item>);

impl_rune_wrapper!(&Arc<dyn Item>, RuneItem);

impl RuneItem {
    #[rune::function]
    fn name(&self) -> Value { self.0.name().clone().wrap() }

    #[rune::function]
    fn category(&self) -> ItemCategory { self.0.category() }

    #[rune::function]
    fn battle_category(&self) -> BattleItemCategory { self.0.battle_category() }

    #[rune::function]
    fn price(&self) -> i32 { self.0.price() }

    #[rune::function]
    fn flags(&self) -> Vec<Value> { self.0.flags().iter().map(|s| s.clone().wrap()).collect() }

    #[rune::function]
    fn has_flag(&self, key: &RuneStringKey) -> bool { self.0.has_flag(&key.0) }
}

#[cfg(test)]
mod tests {
    use super::super::super::tests::{execute_vm, setup_script};
    use super::*;
    use crate::static_data::tests::MockItem;
    use crate::StringKey;

    #[test]
    fn test_get_name() -> anyhow::Result<()> {
        let mut vm = setup_script(
            r#"
            pub fn test_item(item) {
                let name = item.name();
                assert_eq!(name, "Test Item");
            }
            "#,
        )?;
        let mut item = MockItem::new();
        item.expect_name().once().return_const(StringKey::new("Test Item"));
        let item: Arc<dyn Item> = Arc::new(item);
        let wrapped = item.wrap();
        execute_vm!(&mut vm, "test_item", wrapped);
        Ok(())
    }
}
