use crate::script_implementations::wasm::export_registry::wasm_result::{get_value_arc, try_wasm};
use crate::script_implementations::wasm::export_registry::{register, wasm_ok, FunctionEnvMut, WasmResult};
use crate::script_implementations::wasm::extern_ref::ExternRef;
use crate::script_implementations::wasm::script_resolver::WebAssemblyEnv;
use crate::static_data::{Form, StatisticSet};
use crate::StringKey;

register! {

fn form_get_name(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<ExternRef<StringKey>> {
        let form = get_value_arc!(form, env);
        wasm_ok(ExternRef::func_new(&env, form.name().clone().into()))
}

fn form_get_types(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<u64> {
        let form = get_value_arc!(form, env);
        let vec = form.types();
        let wasm_ptr = try_wasm!(env.data().data().copy_value_vec_to_wasm(vec), env);
        let r: u64 = unsafe { std::mem::transmute((wasm_ptr, vec.len() as u32)) };
        wasm_ok(r)
}

fn form_get_height(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<f32> {
        wasm_ok(get_value_arc!(form, env).height())
}

fn form_get_weight(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<f32> {
        wasm_ok(get_value_arc!(form, env).weight())
}

fn form_get_base_experience(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<u32> {
        wasm_ok(get_value_arc!(form, env).base_experience())
}

fn form_get_base_stats(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<ExternRef<StatisticSet<u16>>> {
        let form = get_value_arc!(form, env);
        wasm_ok(ExternRef::func_new(&env, form.base_stats().into()))
}

fn form_get_abilities(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<u64> {
        let form = get_value_arc!(form, env);
        let abilities = form.abilities();
        let mut ability_refs : Vec<u32> = Vec::with_capacity(abilities.len());
        for ability in abilities {
            ability_refs.push(ExternRef::<StringKey>::func_new(&env, ability.clone().into()).index() as u32);
        }
        let wasm_ptr = try_wasm!(env.data().data().copy_value_vec_to_wasm(&ability_refs), env);
        let r: u64 = unsafe { std::mem::transmute((wasm_ptr, ability_refs.len() as u32)) };
        wasm_ok(r)
}

fn form_get_hidden_abilities(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>
) -> WasmResult<u64> {
        let form = get_value_arc!(form, env);
        let abilities = form.hidden_abilities();
        let mut ability_refs : Vec<u32> = Vec::with_capacity(abilities.len());
        for ability in abilities {
            ability_refs.push(ExternRef::<StringKey>::func_new(&env, ability.clone().into()).index() as u32);
        }
        let wasm_ptr = try_wasm!(env.data().data().copy_value_vec_to_wasm(&ability_refs), env);
        let r: u64 = unsafe { std::mem::transmute((wasm_ptr, ability_refs.len() as u32)) };
        wasm_ok(r)
}

fn form_has_flag_by_hash(
    env: FunctionEnvMut<WebAssemblyEnv>,
    form: ExternRef<dyn Form>,
    flag_hash: u32,
) -> WasmResult<u8> {
    let form = get_value_arc!(form, env);
    wasm_ok(if form.has_flag_by_hash(flag_hash) { 1 } else { 0 })
}


}
