use crate::script_implementations::wasm::script_resolver::{WebAssemblyEnv, WebAssemblyEnvironmentData};
use anyhow::Error;
use std::ffi::{c_char, CString};
use std::sync::Arc;
use wasmer::{FromToNativeWasmType, FunctionEnvMut};

/// A result type that can be given to, or returned from WASM.
pub type WasmResult<T> = (u32, T);

/// A result type that can be given to, or returned from WASM, but does not contain a value.
pub type WasmVoidResult = u32;

/// An extension trait for `WasmResult` that provides some convenience methods.
pub(crate) trait WasmVoidResultExtension {
    /// Returns a `WasmVoidResult` that indicates success.
    fn ok() -> Self;
    /// Returns a `WasmVoidResult` that indicates failure.
    fn err(err: anyhow::Error, env: &FunctionEnvMut<WebAssemblyEnv>) -> Self;
    /// Converts a `WasmVoidResult` into a `Result<(), anyhow::Error>`. Used specifically for script
    /// function calls.
    fn into_result_env(self, env: &Arc<WebAssemblyEnvironmentData>) -> anyhow::Result<()>;
}

/// Returns a success `WasmResult` with the given value.
pub(super) fn wasm_ok<T>(value: T) -> WasmResult<T>
where
    T: FromToNativeWasmType,
    <T as FromToNativeWasmType>::Native: Default,
{
    (0, value)
}

/// Returns a failure `WasmResult` with the given error.
#[allow(clippy::unwrap_used)]
pub(super) fn wasm_err<T>(err: Error, env: &FunctionEnvMut<WebAssemblyEnv>) -> WasmResult<T>
where
    T: FromToNativeWasmType + Default,
{
    let s = CString::new(err.to_string()).unwrap();
    let ptr = env.data().data().copy_value_vec_to_wasm(s.as_bytes()).unwrap();
    (ptr, <T as Default>::default())
}

impl WasmVoidResultExtension for WasmVoidResult {
    fn ok() -> Self {
        0
    }

    #[allow(clippy::unwrap_used)]
    fn err(err: Error, env: &FunctionEnvMut<WebAssemblyEnv>) -> Self {
        let s = CString::new(err.to_string()).unwrap();
        let ptr = env.data().data().copy_value_vec_to_wasm(s.as_bytes()).unwrap();
        ptr
    }

    #[allow(clippy::unwrap_used)]
    fn into_result_env(self, env: &Arc<WebAssemblyEnvironmentData>) -> anyhow::Result<()> {
        if self == 0 {
            Ok(())
        } else {
            unsafe {
                let ptr = self;
                let mem: *mut c_char = env.get_raw_pointer(ptr);
                let string = std::ffi::CStr::from_ptr(mem);
                let e = anyhow_ext::anyhow!("{}", string.to_str().unwrap());
                env.script_function_cache().dealloc_cstring(env, ptr).unwrap();
                Err(e)
            }
        }
    }
}

/// Macro to try an expression, and return a `WasmResult` if it fails.
macro_rules! try_wasm {
    ($e:expr, $env:expr) => {
        match $e {
            Ok(v) => v,
            Err(e) => {
                return crate::script_implementations::wasm::export_registry::wasm_err(e.into(), &$env);
            }
        }
    };
}

/// Macro to try an expression, and return a `WasmVoidResult` if it fails.
macro_rules! try_wasm_void {
    ($e:expr, $env:expr) => {
        match $e {
            Ok(v) => v,
            Err(e) => {
                return WasmVoidResult::err(e.into(), &$env);
            }
        }
    };
}

pub(super) use try_wasm;
pub(super) use try_wasm_void;

/// Resolve an externref to a value. Returns an error if this fails.
macro_rules! get_value {
    ($e:expr, $env:expr) => {
        crate::script_implementations::wasm::export_registry::try_wasm!($e.value_func(&$env), $env)
    };
}

/// Resolve an externref of an Arc to a value. Returns an error if this fails.
macro_rules! get_value_arc {
    ($e:expr, $env:expr) => {
        crate::script_implementations::wasm::export_registry::try_wasm!($e.value_func_arc(&$env), $env)
    };
}

/// Resolve an externref to a value. Returns an error if this fails.
macro_rules! get_value_void {
    ($e:expr, $env:expr) => {
        crate::script_implementations::wasm::export_registry::try_wasm_void!($e.value_func(&$env), $env)
    };
}

/// Resolve an externref of an Arc to a value. Returns an error if this fails.
macro_rules! get_value_arc_void {
    ($e:expr, $env:expr) => {
        crate::script_implementations::wasm::export_registry::try_wasm_void!($e.value_func_arc(&$env), $env)
    };
}

pub(super) use get_value;
pub(super) use get_value_arc;
pub(super) use get_value_arc_void;
pub(super) use get_value_void;

/// Resolve an externref to a value, and call a getter on it. Returns an error if this fails.
macro_rules! get_value_call_getter {
    ($e:ident.$func:ident(), $env:expr) => {{
        let _value = crate::script_implementations::wasm::export_registry::try_wasm!($e.value_func(&$env), $env);
        _value.$func()
    }};
}

pub(super) use get_value_call_getter;
