use hashbrown::HashSet;
#[cfg(feature = "serde")] use serde::{Deserialize, Serialize};
use std::any::Any;
use std::fmt::Debug;

use crate::StringKey;

/// An item category defines which bag slot items are stored in.
#[derive(Debug, Copy, Clone)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
#[cfg_attr(feature = "rune", derive(rune::Any))]
#[repr(u8)]
pub enum ItemCategory {
    /// This is where most items should go.
    #[cfg_attr(feature = "rune", rune(constructor))]
    MiscItem,
    /// Pokeballs are used for capturing Pokemons.
    #[cfg_attr(feature = "rune", rune(constructor))]
    Pokeball,
    /// Medicine is used for healing HP, PP, and status effects
    #[cfg_attr(feature = "rune", rune(constructor))]
    Medicine,
    /// Berry is used for all berries.
    #[cfg_attr(feature = "rune", rune(constructor))]
    Berry,
    /// TMHM is used for Technical and Hidden Machines.
    #[cfg_attr(feature = "rune", rune(constructor))]
    TMHM,
    /// Form Changer is used for items that change forms, such as mega stones.
    #[cfg_attr(feature = "rune", rune(constructor))]
    FormChanger,
    /// Key Items are single stored items, generally used for story progression.
    #[cfg_attr(feature = "rune", rune(constructor))]
    KeyItem,
    /// Mail is used for mail items.
    #[cfg_attr(feature = "rune", rune(constructor))]
    Mail,
}

/// A battle item category defines how the item is categorized when in battle.
#[derive(Debug, Copy, Clone)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
#[cfg_attr(feature = "rune", derive(rune::Any))]
#[repr(u8)]
pub enum BattleItemCategory {
    /// This item can't be used in battle.
    #[cfg_attr(feature = "rune", rune(constructor))]
    None,
    /// This item is used for healing Pokemon.
    #[cfg_attr(feature = "rune", rune(constructor))]
    Healing,
    /// This item is used for healing Pokemon from a status.
    #[cfg_attr(feature = "rune", rune(constructor))]
    StatusHealing,
    /// This item is used for capturing Pokemon.
    #[cfg_attr(feature = "rune", rune(constructor))]
    Pokeball,
    /// This item does not belong in above categories, but is still a battle item.
    #[cfg_attr(feature = "rune", rune(constructor))]
    MiscBattleItem,
}

/// An item is an object which the player can pick up, keep in their Bag, and use in some manner
pub trait Item: Debug + Any {
    /// The name of the item.
    fn name(&self) -> &StringKey;
    /// Which bag slot items are stored in.
    fn category(&self) -> ItemCategory;
    /// How the item is categorized when in battle.
    fn battle_category(&self) -> BattleItemCategory;
    /// The buying value of the item.
    fn price(&self) -> i32;
    /// A set of arbitrary flags that can be set on the item.
    fn flags(&self) -> &HashSet<StringKey>;

    /// Checks whether the item has a specific flag.
    fn has_flag(&self, key: &StringKey) -> bool;
}

/// An item is an object which the player can pick up, keep in their Bag, and use in some manner
#[derive(Debug)]
pub struct ItemImpl {
    /// The name of the item.
    name: StringKey,
    /// Which bag slot items are stored in.
    category: ItemCategory,
    /// How the item is categorized when in battle.
    battle_category: BattleItemCategory,
    /// The buying value of the item.
    price: i32,
    /// A set of arbitrary flags that can be set on the item.
    flags: HashSet<StringKey>,
}

impl ItemImpl {
    /// Instantiates an item.
    pub fn new(
        name: &StringKey,
        category: ItemCategory,
        battle_category: BattleItemCategory,
        price: i32,
        flags: HashSet<StringKey>,
    ) -> ItemImpl {
        ItemImpl {
            name: name.clone(),
            category,
            battle_category,
            price,
            flags,
        }
    }
}

impl Item for ItemImpl {
    /// The name of the item.
    fn name(&self) -> &StringKey { &self.name }
    /// Which bag slot items are stored in.
    fn category(&self) -> ItemCategory { self.category }
    /// How the item is categorized when in battle.
    fn battle_category(&self) -> BattleItemCategory { self.battle_category }
    /// The buying value of the item.
    fn price(&self) -> i32 { self.price }
    /// A set of arbitrary flags that can be set on the item.
    fn flags(&self) -> &HashSet<StringKey> { &self.flags }

    /// Checks whether the item has a specific flag.
    fn has_flag(&self, key: &StringKey) -> bool { self.flags.contains(key) }
}

#[cfg(test)]
#[allow(clippy::indexing_slicing)]
pub(crate) mod tests {
    use super::*;

    mockall::mock! {
        #[derive(Debug)]
        pub Item {}
        impl Item for Item {
            fn name(&self) -> &StringKey;
            fn category(&self) -> ItemCategory;
            fn battle_category(&self) -> BattleItemCategory;
            fn price(&self) -> i32;
            fn flags(&self) -> &HashSet<StringKey>;
            fn has_flag(&self, key: &StringKey) -> bool;
        }
    }
}
